# conformal classification
from fortuna.conformal.classification.adaptive_prediction import AdaptivePredictionConformalClassifier
from fortuna.conformal.classification.simple_prediction import SimplePredictionConformalClassifier
from fortuna.conformal.classification.adaptive_conformal_classifier import AdaptiveConformalClassifier
from fortuna.conformal.classification.batch_mvp import BatchMVPConformalClassifier

# conformal regression
from fortuna.conformal.regression.onedim_uncertainty import OneDimensionalUncertaintyConformalRegressor
from fortuna.conformal.regression.quantile import QuantileConformalRegressor
from fortuna.conformal.regression.jackknifeplus import JackknifePlusConformalRegressor
from fortuna.conformal.regression.jackknife_minmax import JackknifeMinmaxConformalRegressor
from fortuna.conformal.regression.cvplus import CVPlusConformalRegressor
from fortuna.conformal.regression.enbpi import EnbPI
from fortuna.conformal.regression.batch_mvp import BatchMVPConformalRegressor
from fortuna.conformal.regression.adaptive_conformal_regressor import AdaptiveConformalRegressor



from __future__ import annotations

import logging
from typing import Optional, Tuple, Dict

import jax.numpy as jnp
from flax.core import FrozenDict
from jax._src.prng import PRNGKeyArray
from jax.flatten_util import ravel_pytree

from fortuna.data.loader import DataLoader, InputsLoader
from fortuna.distribution.gaussian import DiagGaussian
from fortuna.prob_model.fit_config.base import FitConfig
from fortuna.prob_model.joint.base import Joint
from fortuna.prob_model.joint.state import JointState
from fortuna.prob_model.posterior.gaussian_posterior import GaussianPosterior
from fortuna.prob_model.posterior.normalizing_flow.advi import ADVI_NAME
from fortuna.prob_model.posterior.normalizing_flow.advi.advi_approximator import \
    ADVIPosteriorApproximator
from fortuna.prob_model.posterior.normalizing_flow.advi.advi_architecture import \
    ADVIArchitecture
from fortuna.prob_model.posterior.normalizing_flow.advi.advi_state import \
    ADVIState
from fortuna.prob_model.posterior.normalizing_flow.advi.advi_trainer import \
    ADVITrainer, JittedADVITrainer, MultiDeviceADVITrainer
from fortuna.prob_model.posterior.posterior_state_repository import \
    PosteriorStateRepository
from fortuna.typing import Status, Array
from fortuna.utils.device import select_trainer_given_devices
from fortuna.prob_model.posterior.map.map_state import MAPState
from fortuna.utils.nested_dicts import nested_get, nested_unpair, nested_set
from fortuna.prob_model.posterior.map.map_posterior import MAPPosterior
from fortuna.prob_model.posterior.map.map_approximator import MAPPosteriorApproximator
from fortuna.prob_model.posterior.state import PosteriorState
import numpy as np


class ADVIPosterior(GaussianPosterior):
    def __init__(
        self,
        joint: Joint,
        posterior_approximator: ADVIPosteriorApproximator,
    ):
        """
        Automatic Differentiation Variational Inference (ADVI) approximate posterior class.

        Parameters
        ----------
        joint: Joint
            A joint distribution object.
        posterior_approximator: ADVI
            An ADVI posterior approximator.
        """
        super().__init__(joint=joint, posterior_approximator=posterior_approximator)

    def __str__(self):
        return ADVI_NAME

    def fit(
        self,
        train_data_loader: DataLoader,
        val_data_loader: Optional[DataLoader] = None,
        map_fit_config: Optional[FitConfig] = None,
        fit_config: FitConfig = FitConfig(),
        **kwargs,
    ) -> Dict[str, Status]:
        if (
            fit_config.checkpointer.dump_state is True
            and not fit_config.checkpointer.save_checkpoint_dir
        ):
            raise ValueError(
                "`save_checkpoint_dir` must be passed when `dump_state` is set to True."
            )

        log_stds = None

        state, status, n_train_data, n_val_data = self._init_state(
            train_data_loader=train_data_loader,
            val_data_loader=val_data_loader,
            fit_config=fit_config,
            map_fit_config=map_fit_config,
            **kwargs
        )

        rav, self._unravel, self._indices, rav_log_stds = self._get_unravel(state.params, log_stds)

        size_rav = len(rav)
        self.base = DiagGaussian(
            mean=jnp.zeros(size_rav),
            std=jnp.exp(self.posterior_approximator.log_std_base) * jnp.ones(size_rav),
        )
        self.architecture = ADVIArchitecture(
            size_rav, std_init_params=self.posterior_approximator.std_init_params
        )

        trainer_cls = select_trainer_given_devices(
            devices=fit_config.processor.devices,
            BaseTrainer=ADVITrainer,
            JittedTrainer=JittedADVITrainer,
            MultiDeviceTrainer=MultiDeviceADVITrainer,
            disable_jit=fit_config.processor.disable_jit,
        )

        trainer = trainer_cls(
            predict_fn=self.joint.likelihood.prob_output_layer.predict,
            save_checkpoint_dir=fit_config.checkpointer.save_checkpoint_dir,
            save_every_n_steps=fit_config.checkpointer.save_every_n_steps,
            keep_top_n_checkpoints=fit_config.checkpointer.keep_top_n_checkpoints,
            disable_training_metrics_computation=fit_config.monitor.disable_training_metrics_computation,
            eval_every_n_epochs=fit_config.monitor.eval_every_n_epochs,
            early_stopping_monitor=fit_config.monitor.early_stopping_monitor,
            early_stopping_min_delta=fit_config.monitor.early_stopping_min_delta,
            early_stopping_patience=fit_config.monitor.early_stopping_patience,
            base=self.base,
            architecture=self.architecture,
            which_params=self.posterior_approximator.which_params,
            all_params=state.params if self.posterior_approximator.which_params else None,
            indices=self._indices,
            unravel=self._unravel
        )

        state = ADVIState.init(
            params=FrozenDict(
                    trainer.init_params(
                        self.rng.get(),
                        mean=rav,
                        log_std=None if log_stds is None else (rav, rav_log_stds)
                    ),
                ),
            mutable=getattr(state, "mutable", state.mutable),
            optimizer=fit_config.optimizer.method,
            calib_params=getattr(state, "calib_params", state.calib_params),
            calib_mutable=getattr(state, "calib_mutable", state.calib_mutable),
        )

        logging.info("Run ADVI.")
        state, status["advi"] = trainer.train(
            rng=self.rng.get(),
            state=state,
            loss_fun=self.joint._batched_negative_log_joint_prob,
            training_dataloader=train_data_loader,
            training_dataset_size=n_train_data,
            n_epochs=fit_config.optimizer.n_epochs,
            metrics=fit_config.monitor.metrics,
            validation_dataloader=val_data_loader,
            validation_dataset_size=n_val_data,
            verbose=fit_config.monitor.verbose,
            unravel=self._unravel,
            n_samples=self.posterior_approximator.n_loss_samples,
            callbacks=fit_config.callbacks,
        )
        trainer._all_params = None

        self.state = PosteriorStateRepository(
            fit_config.checkpointer.save_checkpoint_dir
            if fit_config.checkpointer.dump_state is True
            else None
        )
        self.state.put(state, keep=fit_config.checkpointer.keep_top_n_checkpoints)
        logging.info("Fit completed.")
        return status

    def sample(
        self,
        rng: Optional[PRNGKeyArray] = None,
        input_shape: Optional[Tuple[int, ...]] = None,
        inputs_loader: Optional[InputsLoader] = None,
        inputs: Optional[Array] = None,
        **kwargs,
    ) -> JointState:
        """
        Sample from the posterior distribution. Either `input_shape` or `_inputs_loader` must be passed.
        Parameters
        ----------
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        input_shape: Optional[Tuple[int, ...]]
            Shape of a single input.
        inputs_loader: Optional[InputsLoader]
            Input data loader. If `input_shape` is passed, then `inputs` and `inputs_loader` are ignored.
        inputs: Optional[Array]
            Input variables.
        Returns
        -------
        JointState
            A sample from the posterior distribution.
        """
        if rng is None:
            rng = self.rng.get()
        state = self.state.get()

        if not hasattr(self, "base") or not hasattr(self, "_unravel"):
            if self.posterior_approximator.which_params is None:
                n_params = len(ravel_pytree(state.params)[0]) // 2
            else:
                n_params = len(ravel_pytree(
                    nested_unpair(
                        d=state.params.unfreeze(),
                        key_paths=self.posterior_approximator.which_params,
                        labels=("mean", "log_std")
                    )[0]
                )[0])
            self.base = DiagGaussian(
                mean=jnp.zeros(n_params),
                std=jnp.exp(self.posterior_approximator.log_std_base) * jnp.ones(n_params),
            )
            self.architecture = ADVIArchitecture(
                n_params, std_init_params=self.posterior_approximator.std_init_params
            )
            self._unravel, self._indices = self._get_unravel(state.params)[1:2]

        if self.posterior_approximator.which_params is None:
            means = self._unravel(
                self.architecture.forward(
                    {s: ravel_pytree({k: v["params"][s] for k, v in state.params.items()})[0] for s in ["mean", "log_std"]},
                    self.base.sample(rng)
                )[0][0]
            )
        else:
            means, log_stds = nested_unpair(
                        d=state.params.unfreeze(),
                        key_paths=self.posterior_approximator.which_params,
                        labels=("mean", "log_std")
                    )
            rav_params = {
                k: ravel_pytree(
                    [nested_get(
                        d=d,
                        keys=path
                    ) for path in self.posterior_approximator.which_params]
                )[0] for k, d in zip(
                    ["mean", "log_std"],
                    [means, log_stds]
                )
            }
            rav_params = self.architecture.forward(
                params=rav_params,
                u=self.base.sample(rng)
            )[0][0]

            means = FrozenDict(
                nested_set(
                    d=means,
                    key_paths=self.posterior_approximator.which_params,
                    objs=tuple([_unravel(rav_params[self._indices[i]:self._indices[i + 1]]) for i, _unravel in enumerate(self._unravel)]),
                )
            )

        return JointState(
            params=FrozenDict(means),
            mutable=state.mutable,
            calib_params=state.calib_params,
            calib_mutable=state.calib_mutable,
        )

    def _init_state(
            self,
            train_data_loader: DataLoader,
            val_data_loader: DataLoader,
            fit_config: FitConfig,
            map_fit_config: Optional[FitConfig] = None,
            **kwargs
    ) -> Tuple[PosteriorState, Dict[str, Status], int, Optional[int]]:
        allowed_states = (MAPState, ADVIState)
        state = None
        status = dict()

        if fit_config.checkpointer.restore_checkpoint_path is not None:
            state = self.restore_checkpoint(
                restore_checkpoint_path=fit_config.checkpointer.restore_checkpoint_path,
                optimizer=fit_config.optimizer.method,
            )

            if not isinstance(state, allowed_states):
                raise ValueError(f"The type of the restored checkpoint must be within {allowed_states}. "
                                 f"However, {fit_config.checkpointer.restore_checkpoint_path} pointed to a state "
                                 f"with type {type(state)}.")

            if isinstance(state, ADVIState):
                means, log_stds = nested_unpair(
                    d=state.params.unfreeze(),
                    key_paths=self.posterior_approximator.which_params,
                    labels=("mean", "log_std")
                ) if self.posterior_approximator.which_params is not None else \
                    [FrozenDict({k: dict(params=v["params"][s]) for k, v in state.params.items()}) for s in ["mean", "log_std"]]
                means, log_stds = FrozenDict(means), FrozenDict(log_stds)
                state = state.replace(params=means)
                del means
        elif map_fit_config is not None:
            map_posterior = MAPPosterior(
                self.joint, posterior_approximator=MAPPosteriorApproximator()
            )
            map_posterior.rng = self.rng
            logging.info("Do a preliminary run of MAP.")
            status["map"] = map_posterior.fit(
                rng=self.rng.get(),
                train_data_loader=train_data_loader,
                val_data_loader=val_data_loader,
                fit_config=map_fit_config,
                **kwargs,
            )
            logging.info("Preliminary run with MAP completed.")
            state = map_posterior.state.get()
        elif self.posterior_approximator.which_params is not None:
            logging.warning(
                f"ADVI will run exclusively over the parameters specified in `which_params`."
                f"All the other parameters will be initialized randomly. "
                f"If you want to train the other parameters too, "
                f"and yet run ADVI only on the parameters in `which_params`, "
                f"you can either restore an existing checkpoint, or set `map_fit_config`. "
                f"In the latter case, a pre-run of MAP will run before ADVI.")

        state, n_train_data, n_val_data = self._init(
            train_data_loader, val_data_loader, state
        )

        return state, status, n_train_data, n_val_data

    def _get_unravel(self, params, log_stds: Optional[Array] = None):
        if self.posterior_approximator.which_params is None:
            rav, unravel = ravel_pytree(params)
            rav_log_stds = ravel_pytree(log_stds)[0] if log_stds is not None else None
            indices = None
        else:
            def unravel_fn(_params, _path):
                return ravel_pytree(nested_get(_params, _path))
            rav, unravel, indices, rav_log_stds = [], [], [], []
            for path in self.posterior_approximator.which_params:
                _rav, _unravel = unravel_fn(params, path)
                unravel.append(_unravel)
                rav.append(_rav)
                if log_stds is not None:
                    rav_log_stds.append(unravel_fn(log_stds, path)[0])
                indices.append(len(_rav))
            rav = jnp.concatenate(rav)
            if log_stds is not None:
                rav_log_stds = jnp.concatenate(rav_log_stds)
            else:
                rav_log_stds = None
            unravel = tuple(unravel)
            indices = np.concatenate((np.array([0]), np.cumsum(indices)))
        return rav, unravel, indices, rav_log_stds



from fortuna.typing import Array
import jax.numpy as jnp


def scaled_mse_fn(
        outputs: Array,
        targets: Array,
) -> jnp.ndarray:
    means, log_vars = jnp.split(outputs, 2, axis=-1)
    return jnp.mean(jnp.sum(jnp.exp(-log_vars) * (targets - means) ** 2 + log_vars, -1))



from fortuna.prob_model.classification import ProbClassifier
from fortuna.prob_model.regression import ProbRegressor

from fortuna.prob_model.posterior.deep_ensemble.deep_ensemble_approximator import \
    DeepEnsemblePosteriorApproximator
from fortuna.prob_model.posterior.laplace.laplace_approximator import \
    LaplacePosteriorApproximator
from fortuna.prob_model.posterior.map.map_posterior import \
    MAPPosteriorApproximator
from fortuna.prob_model.posterior.normalizing_flow.advi.advi_approximator import \
    ADVIPosteriorApproximator
from fortuna.prob_model.posterior.swag.swag_approximator import \
    SWAGPosteriorApproximator
from fortuna.prob_model.posterior.sngp.sngp_approximator import \
    SNGPPosteriorApproximator

from fortuna.prob_model.calib_config.base import CalibConfig
from fortuna.prob_model.calib_config.checkpointer import CalibCheckpointer
from fortuna.prob_model.calib_config.monitor import CalibMonitor
from fortuna.prob_model.calib_config.optimizer import CalibOptimizer
from fortuna.prob_model.calib_config.processor import CalibProcessor

from fortuna.prob_model.fit_config.base import FitConfig
from fortuna.prob_model.fit_config.checkpointer import FitCheckpointer
from fortuna.prob_model.fit_config.monitor import FitMonitor
from fortuna.prob_model.fit_config.optimizer import FitOptimizer
from fortuna.prob_model.fit_config.processor import FitProcessor

from fortuna.prob_model.fit_config.processor import FitProcessor



from __future__ import annotations

import abc
import logging
from typing import Callable, Union, List, Tuple
from fortuna.typing import Array
from fortuna.data.loader import DataLoader, InputsLoader
import jax.numpy as jnp
from jax import vmap


class Group:
    def __init__(self, group_fn: Callable[[Array], Array]):
        self.group_fn = group_fn

    def __call__(self, x):
        g = self.group_fn(x)
        if g.ndim > 1:
            raise ValueError("Evaluations of the group function `group_fn` must be one-dimensional arrays.")
        if jnp.any((g != 0) * (g != 1)):
            raise ValueError("The group function `threshold_fn` must take values in {0, 1}.")
        return g.astype(bool)


class Normalizer:
    def __init__(self, xmin: Array, xmax: Array):
        self.xmin = xmin
        self.xmax = xmax if xmax != xmin else xmin + 1

    def normalize(self, x: Array) -> Array:
        return (x - self.xmin) / (self.xmax - self.xmin)

    def unnormalize(self, y: Array) -> Array:
        return self.xmin + (self.xmax - self.xmin) * y


class Score:
    def __init__(self, score_fn: Callable[[Array, Array], Array]):
        self.score_fn = score_fn

    def __call__(self, x: Array, y: Array):
        s = self.score_fn(x, y)
        if s.ndim > 1:
            raise ValueError("Evaluations of the score function `score_fn` must be one-dimensional arrays, "
                             f"but its shape was {s.shape}.")
        return s


class BatchMVPConformalMethod(abc.ABC):
    def __init__(
            self,
            score_fn: Callable[[Array, Array], Array],
            group_fns: List[Callable[[Array], Array]],
            n_buckets: int = 100
    ):
        super().__init__()
        self.score_fn = Score(score_fn)
        self.group_fns = [Group(g) for g in group_fns]
        self.buckets = jnp.linspace(0, 1, n_buckets + 1)
        self.n_buckets = n_buckets + 1

    def threshold_score(
            self,
            val_data_loader: DataLoader,
            test_inputs_loader: InputsLoader,
            error: float = 0.05,
            tol: float = 1e-4,
            n_rounds: int = 1000,
            return_max_calib_error: bool = False
    ) -> Union[Array, Tuple[Array, List[Array]]]:
        """
        Compute a threshold :math:`f(x)` of the score functions :math:`s(x,y)` for each test input :math:`x`.
        Given these threshold, conformal sets can be formulated as :math:`C(x) = \{y: s(x,y) \le f(x)\}`.

        Parameters
        ----------
        val_data_loader: DataLoader
            A data loader of validation data.
        test_inputs_loader: InputsLoader
            A loader of test input data points.
        error: float
            A desired coverage error.
        tol: float
            A tolerance for the maximum calibration error.
        n_rounds: int
            The maximum number of updates the algorithm will run for.
        return_max_calib_error: bool
            Whether to return a list of computed maximum calibration error, that is the larger calibration error
            over the different groups.

        Returns
        -------
        Union[Array, Tuple[Array, List[Array]]]
            The compute threshold of the score function for each test input.
        """
        quantile = 1 - error

        scores, thresholds, groups = [], [], []
        for inputs, targets in val_data_loader:
            scores.append(self.score_fn(inputs, targets))
            thresholds.append(jnp.zeros(inputs.shape[0]))
            groups.append(jnp.concatenate([g(inputs)[:, None] for g in self.group_fns], axis=1))
        scores, thresholds, groups = jnp.concatenate(scores), jnp.concatenate(thresholds), jnp.concatenate(groups, 0)

        test_thresholds, test_groups = [], []
        for inputs in test_inputs_loader:
            test_thresholds.append(jnp.zeros(inputs.shape[0]))
            test_groups.append(jnp.concatenate([g(inputs)[:, None] for g in self.group_fns], axis=-1))
        test_thresholds, test_groups = jnp.concatenate(test_thresholds), jnp.concatenate(test_groups, 0)

        normalizer = Normalizer(jnp.min(scores), jnp.max(scores))
        scores = normalizer.normalize(scores)

        n_groups = groups.shape[1]

        def compute_probability_error(v: Array, g: Array, delta: Union[Array, float] = 0.):
            b = (jnp.abs(thresholds - v) < 0.5 / self.n_buckets) * groups[:, g]
            filtered_scores = jnp.where(b, scores, -jnp.ones_like(scores))
            conds = (filtered_scores <= v + delta) * (filtered_scores != -1)
            prob_b = jnp.mean(b)
            prob = jnp.where(prob_b > 0, jnp.mean(conds) / prob_b, 0.)
            return (quantile - prob) ** 2

        def calibration_error(v, g):
            b = (jnp.abs(thresholds - v) < 0.5 / self.n_buckets) * groups[:, g]
            filtered_scores = jnp.where(b, scores, -jnp.ones_like(scores))
            conds = (filtered_scores <= v) * (filtered_scores != -1)
            prob_b = jnp.mean(b)
            prob = jnp.where(prob_b > 0, jnp.mean(conds) / prob_b, 0.)
            return prob_b * (quantile - prob) ** 2

        max_calib_errors = None
        if return_max_calib_error:
            max_calib_errors = []

        for t in range(n_rounds):
            calib_error_vg = vmap(lambda g: vmap(lambda v: calibration_error(v, g))(self.buckets))(jnp.arange(n_groups))
            max_calib_error = calib_error_vg.sum(1).max()
            if return_max_calib_error:
                max_calib_errors.append(max_calib_error)
            if max_calib_error <= tol:
                logging.info(f"The algorithm produced a {tol}-approximately {quantile}-quantile multicalibrated "
                             f"threshold function after {t} rounds.")
                break

            gt, idx_vt = jnp.unravel_index(jnp.argmax(calib_error_vg), (n_groups, self.n_buckets))
            vt = self.buckets[idx_vt]

            deltat = self.buckets[
                jnp.argmin(
                    jnp.abs(vmap(lambda delta: compute_probability_error(vt, gt, delta))(self.buckets))
                )
            ]
            bt = (jnp.abs(thresholds - vt) < 0.5 / self.n_buckets) * groups[:, gt]
            thresholds = thresholds.at[bt].set(jnp.minimum(thresholds[bt] + deltat, jnp.ones_like(thresholds[bt])))
            test_bt = (jnp.abs(test_thresholds - vt) < 0.5 / self.n_buckets) * test_groups[:, gt]
            test_thresholds = test_thresholds.at[test_bt].set(
                jnp.minimum(test_thresholds[test_bt] + deltat, jnp.ones_like(test_thresholds[test_bt]))
            )

        test_thresholds = normalizer.unnormalize(test_thresholds)
        if return_max_calib_error:
            return test_thresholds, max_calib_errors
        return test_thresholds



import abc

from fortuna.prob_model.posterior.base import Posterior
from typing import Optional
from flax.core import FrozenDict
from jax._src.prng import PRNGKeyArray
from fortuna.prob_model.joint.state import JointState
from fortuna.utils.nested_dicts import nested_set, nested_unpair, nested_get
from jax.tree_util import tree_map
from fortuna.utils.random import generate_random_normal_like_tree


class GaussianPosterior(Posterior, abc.ABC):
    def _sample_diag_gaussian_from_mean_and_std(self, rng: Optional[PRNGKeyArray] = None) -> JointState:
        if rng is None:
            rng = self.rng.get()
        state = self.state.get()

        if hasattr(self.posterior_approximator, "which_params") and self.posterior_approximator.which_params is not None:
            mean, std = nested_unpair(
                state.params.unfreeze(),
                self.posterior_approximator.which_params,
                ("mean", "std"),
            )

            noise = generate_random_normal_like_tree(rng, std)
            params = nested_set(
                d=mean,
                key_paths=self.posterior_approximator.which_params,
                objs=tuple(
                    [
                        tree_map(
                            lambda m, s, e: m + s * e,
                            nested_get(mean, keys),
                            nested_get(std, keys),
                            nested_get(noise, keys),
                        )
                        for keys in self.posterior_approximator.which_params
                    ]
                ),
            )
            for k, v in params.items():
                params[k] = FrozenDict(v)
            state = state.replace(params=FrozenDict(params))
        else:
            mean, std = dict(), dict()
            for k, v in state.params.items():
                mean[k] = FrozenDict({"params": v["params"]["mean"]})
                std[k] = FrozenDict({"params": v["params"]["std"]})

            state = state.replace(
                params=FrozenDict(
                    tree_map(
                        lambda m, s, e: m + s * e,
                        mean,
                        std,
                        generate_random_normal_like_tree(rng, std),
                    )
                )
            )

        return JointState(
            params=state.params,
            mutable=state.mutable,
            calib_params=state.calib_params,
            calib_mutable=state.calib_mutable,
        )



import abc
import logging
from typing import Callable, Optional, Tuple

import jax.numpy as jnp
import optax
from flax.traverse_util import path_aware_map

from fortuna.data.loader import DataLoader
from fortuna.typing import Path, Status, Predictions, Targets, Uncertainties, Outputs
from fortuna.utils.device import select_trainer_given_devices
from fortuna.utils.random import RandomNumberGenerator
from fortuna.calib_model.calib_state_repository import CalibStateRepository
from fortuna.calib_model.calib_mixin import WithCalibCheckpointingMixin
from fortuna.calib_model.calib_model_calibrator import CalibModelCalibrator, \
    JittedCalibModelCalibrator, MultiDeviceCalibModelCalibrator
from fortuna.calib_model.config.base import Config
from fortuna.calib_model.loss import Loss
from fortuna.model.model_manager.state import ModelManagerState
from fortuna.calib_model.state import CalibState
from flax.core.frozen_dict import freeze


class CalibModel(WithCalibCheckpointingMixin, abc.ABC):
    def __init__(
            self,
            seed: int = 0
    ):
        """
        A calibration model.

        Parameters
        ----------
        seed: int = 0
            Random seed.
        """
        super().__init__()
        self.rng = RandomNumberGenerator(seed=seed)
        self.__set_rng()

    def __set_rng(self):
        self.model_manager.rng = self.rng
        self.prob_output_layer.rng = self.rng
        self.likelihood.rng = self.rng
        self.predictive.rng = self.rng

    def _calibrate(
        self,
        calib_data_loader: DataLoader,
        uncertainty_fn: Callable[[Predictions, Uncertainties, Targets], jnp.ndarray],
        loss_fn: Callable[[Outputs, Targets], jnp.ndarray],
        val_data_loader: Optional[DataLoader] = None,
        config: Config = Config()
    ) -> Status:
        if (
            config.checkpointer.dump_state is True
            and not config.checkpointer.save_checkpoint_dir
        ):
            raise ValueError(
                "`save_checkpoint_dir` must be passed when `dump_state` is set to True."
            )

        trainer_cls = select_trainer_given_devices(
            devices=config.processor.devices,
            BaseTrainer=CalibModelCalibrator,
            JittedTrainer=JittedCalibModelCalibrator,
            MultiDeviceTrainer=MultiDeviceCalibModelCalibrator,
            disable_jit=config.processor.disable_jit,
        )

        trainer = trainer_cls(
            predict_fn=self.prob_output_layer.predict,
            uncertainty_fn=uncertainty_fn,
            save_checkpoint_dir=config.checkpointer.save_checkpoint_dir,
            save_every_n_steps=config.checkpointer.save_every_n_steps,
            keep_top_n_checkpoints=config.checkpointer.keep_top_n_checkpoints,
            disable_training_metrics_computation=config.monitor.disable_calibration_metrics_computation,
            eval_every_n_epochs=config.monitor.eval_every_n_epochs,
            early_stopping_monitor=config.monitor.early_stopping_monitor,
            early_stopping_min_delta=config.monitor.early_stopping_min_delta,
            early_stopping_patience=config.monitor.early_stopping_patience,
        )

        state = self._init_state(calib_data_loader, config)

        if config.optimizer.freeze_map is not None:
            partition_optimizers = {"trainable": config.optimizer.method, "frozen": optax.set_to_zero()}
            partition_params = freeze(path_aware_map(config.optimizer.freeze_map, state.params))
            config.optimizer.method = optax.multi_transform(partition_optimizers, partition_params)
            state = self._init_state(calib_data_loader, config)

        loss = Loss(self.likelihood, loss_fn=loss_fn)
        loss.rng = self.rng

        n_calib_data = calib_data_loader.size
        n_val_data = val_data_loader.size if val_data_loader is not None else None

        if config.monitor.verbose:
            logging.info("Start calibration.")

        state, status = trainer.train(
            rng=self.rng.get(),
            state=state,
            loss_fun=loss,
            training_dataloader=calib_data_loader,
            training_dataset_size=n_calib_data,
            n_epochs=config.optimizer.n_epochs,
            metrics=config.monitor.metrics,
            validation_dataloader=val_data_loader,
            validation_dataset_size=n_val_data,
            verbose=config.monitor.verbose,
            callbacks=config.callbacks,
        )
        self.predictive.state = CalibStateRepository(
            config.checkpointer.save_checkpoint_dir
            if config.checkpointer.dump_state is True
            else None
        )
        self.predictive.state.put(state, keep=config.checkpointer.keep_top_n_checkpoints)
        if config.monitor.verbose:
            logging.info("Calibration completed.")
        return status

    def load_state(self, checkpoint_path: Path) -> None:
        """
        Load the state of the posterior distribution from a checkpoint path. The checkpoint must be compatible with the
        probabilistic model.

        Parameters
        ----------
        checkpoint_path : Path
            Path to a checkpoint file or directory to restore.
        """
        try:
            self.restore_checkpoint(checkpoint_path)
        except ValueError:
            raise ValueError(
                f"No checkpoint was found in `checkpoint_path={checkpoint_path}`."
            )
        self.predictive.state = CalibStateRepository(checkpoint_dir=checkpoint_path)

    def save_state(
        self, checkpoint_path: Path, keep_top_n_checkpoints: int = 1
    ) -> None:
        return self.predictive.state.put(
            self.predictive.state.get(),
            checkpoint_path=checkpoint_path,
            keep=keep_top_n_checkpoints,
        )

    def _get_output_dim(self, input_shape: Tuple, **kwargs) -> int:
        """
        Initialize the state of the joint distribution.

        Parameters
        ----------
        input_shape : Tuple
            The shape of the input variable.

        Returns
        -------
        A state of the joint distribution.
        """
        oms = ModelManagerState.init_from_dict(
            self.model_manager.init(
                input_shape, rng=self.rng.get(), **kwargs
            )
        )
        return self.model_manager.apply(
            oms.params, jnp.zeros((1,) + input_shape), mutable=oms.mutable
        ).shape[-1]

    def _init(self, data_loader: DataLoader, config: Config):
        for inputs, targets in data_loader:
            input_shape = inputs.shape[1:]
            break

        state = ModelManagerState.init_from_dict(self.likelihood.model_manager.init(input_shape, rng=self.rng.get()))
        return CalibState.init(params=state.params, mutable=state.mutable, optimizer=config.optimizer.method)

    def _init_state(self, calib_data_loader: DataLoader, config: Config) -> CalibState:
        if config.checkpointer.restore_checkpoint_path is None:
            if config.checkpointer.start_from_current_state:
                state = self.predictive.state.get(optimizer=config.optimizer.method)
            else:
                state = self._init(calib_data_loader, config)
        else:
            if config.checkpointer.start_from_current_state:
                logging.warning("`config.checkpointer.start_from_current_state` will be ignored since "
                                "`config.checkpointer.restore_checkpoint_path` is given.")
            state = self.restore_checkpoint(
                restore_checkpoint_path=config.checkpointer.restore_checkpoint_path,
                optimizer=config.optimizer.method,
            )
        return state



from fortuna.calib_model.base import CalibModel
from fortuna.calib_model.predictive.classification import ClassificationPredictive
from fortuna.prob_output_layer.classification import ClassificationProbOutputLayer
from fortuna.model.model_manager.classification import ClassificationModelManager
from fortuna.likelihood.classification import ClassificationLikelihood
from fortuna.typing import Status, Outputs, Targets
from flax import linen as nn
from fortuna.data import DataLoader
from fortuna.calib_model.config.base import Config
from fortuna.loss.classification.focal_loss import focal_loss_fn
from typing import Optional, Callable
import numpy as np
import jax.numpy as jnp


class CalibClassifier(CalibModel):
    def __init__(
            self,
            model: nn.Module,
            seed: int = 0):
        r"""
        A calibration classifier class.

        Parameters
        ----------
        model : nn.Module
            A model describing the deterministic relation between inputs and outputs. The outputs must correspond to
            the logits of a softmax probability vector. The output dimension must be the same as the number of classes.
            Let :math:`x` be input variables and :math:`w` the random model parameters. Then the model is described by
            a function :math:`f(w, x)`, where each component of :math:`f` corresponds to one of the classes.
        seed: int
            A random seed.

        Attributes
        ----------
        model : nn.Module
            See `model` in `Parameters`.
        model_manager : ClassificationModelManager
            This object orchestrates the model's forward pass.
        prob_output_layer : ClassificationProbOutputLayer
            This object characterizes the distribution of target variable given the outputs. It is defined
            by :math:`p(y|o)=\text{Categorical}(y|p=softmax(o))`,
            where :math:`o` denotes the outputs and :math:`y` denotes a target variable.
        likelihood : ClassificationLikelihood
            The likelihood function. This is defined by
            :math:`p(y|w, \phi, x) = \text{Categorical}(y|p=\text{softmax}(g(\phi, f(w, x)))`.
        predictive : ClassificationPredictive
            This denotes the predictive distribution, that is :math:`p(y|x, \mathcal{D})`.
        """
        self.model_manager = ClassificationModelManager(model)
        self.prob_output_layer = ClassificationProbOutputLayer()
        self.likelihood = ClassificationLikelihood(
                model_manager=self.model_manager,
                prob_output_layer=self.prob_output_layer,
                output_calib_manager=None
            )
        self.predictive = ClassificationPredictive(
            likelihood=self.likelihood
        )
        super().__init__(seed=seed)

    def _check_output_dim(self, data_loader: DataLoader):
        if data_loader.size == 0:
            raise ValueError(
                """`data_loader` is either empty or incorrectly constructed."""
            )
        data_output_dim = len(np.unique(data_loader.to_array_targets()))
        for x, y in data_loader:
            input_shape = x.shape[1:]
            break
        model_manager_output_dim = self._get_output_dim(input_shape)
        if model_manager_output_dim != data_output_dim:
            raise ValueError(
                f"""The outputs dimension of `model` must correspond to the number of different classes
            in the target variables of `_data_loader`. However, {model_manager_output_dim} and {data_output_dim} were 
            found, respectively."""
            )

    def calibrate(
        self,
        calib_data_loader: DataLoader,
        val_data_loader: Optional[DataLoader] = None,
        loss_fn: Callable[[Outputs, Targets], jnp.ndarray] = focal_loss_fn,
        config: Config = Config()
    ) -> Status:
        """
        Calibrate the calibration classifier.

        Parameters
        ----------
        calib_data_loader : DataLoader
            A calibration data loader.
        val_data_loader : DataLoader
            A validation data loader.
        loss_fn: Callable[[Outputs, Targets], jnp.ndarray]
            The loss function to use for calibration.
        config : Config
            An object to configure the calibration.

        Returns
        -------
        Status
            A calibration status object. It provides information about the calibration.
        """
        self._check_output_dim(calib_data_loader)
        if val_data_loader is not None:
            self._check_output_dim(val_data_loader)
        return self._calibrate(
            calib_data_loader=calib_data_loader,
            uncertainty_fn=config.monitor.uncertainty_fn if config.monitor.uncertainty_fn is not None
            else self.prob_output_layer.mean,
            val_data_loader=val_data_loader,
            loss_fn=loss_fn,
            config=config,
        )



from fortuna.model.constant import ConstantModel
from fortuna.model.hyper import HyperparameterModel
from fortuna.model.lenet import LeNet5
from fortuna.model.linear import LinearModel
from fortuna.model.mlp import MLP
from fortuna.model.resnet import (ResNet18, ResNet34, ResNet50, ResNet101,
                                  ResNet152, ResNet200)
from fortuna.model.scalar_constant import ScalarConstantModel
from fortuna.model.scalar_hyper import ScalarHyperparameterModel
from fortuna.model.wideresnet import WideResNet28_10



from fortuna.prob_model.posterior.base import PosteriorApproximator
from fortuna.prob_model.posterior.normalizing_flow.advi import ADVI_NAME
from typing import Optional, Tuple, List


class ADVIPosteriorApproximator(PosteriorApproximator):
    def __init__(
        self,
        std_init_params: float = 0.1,
        log_std_base: float = -2.3,
        n_loss_samples: int = 3,
        which_params: Optional[Tuple[List, ...]] = None
    ):
        """
        Automatic Differentiation Variational Inference (ADVI) approximator. It is responsible to define how the
        posterior distribution is approximated.

        Parameters
        ----------
        std_init_params : float
            The standard deviation of the Gaussian distribution used to initialize the parameters of the flow.
        log_std_base : float
            The normalizing flow transforms a base distribution into an approximation of the posterior. The base
            distribution is assumed to be an isotropic Gaussian, with this argument as the log-standard deviation.
        n_loss_samples : int
            Number of samples to approximate the loss, that is the KL divergence (or the ELBO, equivalently).
        which_params: Optional[Tuple[List, ...]]
            Sequences of keys to the parameters of the probabilistic model for which to define the Laplace
            approximation. If `which_params` is not available, the posterior approximation will be over all parameters.
        """
        self.std_init_params = std_init_params
        self.log_std_base = log_std_base
        self.n_loss_samples = n_loss_samples

        if which_params:
            if not isinstance(which_params, tuple):
                raise ValueError("`which_params` must be a tuple of lists.")
            for list_keys in which_params:
                if not isinstance(list_keys, list):
                    raise ValueError("Each element in `which_params` must be a list.")
        self.which_params = which_params

    def __str__(self):
        return ADVI_NAME



"""
The code has been taken from https://github.com/google/edward2/blob/main/edward2/jax/nn/normalization.py
"""
import dataclasses
from typing import Any, Callable, Mapping, Optional, Tuple, Type

from jax.random import PRNGKeyArray
import flax.core
import flax.linen as nn
import jax
import jax.numpy as jnp
import numpy as np

from fortuna.typing import Shape, Array


def _l2_normalize(x: Array, eps: float = 1e-12) -> Array:
    return x * jax.lax.rsqrt(jnp.maximum(jnp.square(x).sum(), eps))


class SpectralNormalization(nn.Module):
    """
    Implements spectral normalization for linear layers.

    See `Spectral Normalization for Generative Adversarial Networks <https://arxiv.org/abs/1802.05957>`_ .

    Attributes
    ----------
    layer: nn.Module
        A Flax layer to apply normalization to.
    iteration: int
        The number of power iterations to estimate weight matrix's singular value.
    norm_multiplier: float
        Multiplicative constant to threshold the normalization.
        Usually under normalization, the singular value will converge to this value.
    u_init: Callable[[PRNGKeyArray, Shape, Type], Array]
        Initializer function for the first left singular vectors of the kernel.
    v_init: Callable[[PRNGKeyArray, Shape, Type], Array]
        Initializer function for the first right singular vectors of the kernel.
    kernel_apply_kwargs: Optional[Mapping[str, Any]]
        Updated keyword arguments to clone the input layer.
        The cloned layer represents the linear operator performed by the weight matrix.
        If not specified, that operator follows SN-GAN implementation
        (`Takeru M. et al <https://arxiv.org/abs/1802.05957>`_).
        In particular, for Dense layers the default behavior is equivalent to using a cloned layer with no bias
        (by specifying `kernel_apply_kwargs=dict(use_bias=False)`). With this customization, we
        can have the same implementation (inspried by
        (`Stephan H., 2020 <https://nbviewer.jupyter.org/gist/shoyer/fa9a29fd0880e2e033d7696585978bfc>`_)
        for different interpretations of Conv layers. Also see `SpectralNormalizationConv2D` for
        an example of using this attribute.
    kernel_name: str
        Name of the kernel parameter of the input layer.
    layer_name: Optional[str]
        Name of the input layer
    update_singular_value_estimate: Optional[bool]
        Whether to perform power interations to update the singular value estimate.
    """
    layer: nn.Module
    iteration: int = 1
    norm_multiplier: float = 0.95
    u_init: Callable[[PRNGKeyArray, Shape, Type], Array] = nn.initializers.normal(
        stddev=0.05
    )
    v_init: Callable[[PRNGKeyArray, Shape, Type], Array] = nn.initializers.normal(
        stddev=0.05
    )
    kernel_apply_kwargs: Optional[Mapping[str, Any]] = None
    kernel_name: str = "kernel"
    layer_name: Optional[str] = None
    update_singular_value_estimate: Optional[bool] = None

    def _get_singular_vectors(
        self,
        initializing: bool,
        kernel_apply: Callable,
        in_shape: Shape,
        dtype: Type
    ) -> Tuple[nn.Variable, nn.Variable]:
        if initializing:
            rng_u = self.make_rng("params")
            rng_v = self.make_rng("params")
            # Interpret output shape (not that this does not cost any FLOPs).
            out_shape = jax.eval_shape(
                kernel_apply, jax.ShapeDtypeStruct(in_shape, dtype)
            ).shape
        else:
            rng_u = rng_v = out_shape = None
        u = self.variable("spectral_stats", "u", self.u_init, rng_u, out_shape, dtype)
        v = self.variable("spectral_stats", "v", self.v_init, rng_v, in_shape, dtype)
        return u, v

    @nn.compact
    def __call__(self, inputs: Array, update_singular_value_estimate: Optional[bool] = None) -> Array:
        """
        Applies a linear transformation with spectral normalization to the inputs.

        Parameters
        ----------
        inputs: Array
            The nd-array to be transformed.
        update_singular_value_estimate: Optional[bool]
            Whether to perform power interations to update the singular value estimate.

        Returns
        -------
        Array
            The transformed input.
        """
        update_singular_value_estimate = nn.merge_param(
            'update_singular_value_estimate', self.update_singular_value_estimate, update_singular_value_estimate
        )
        layer_name = (
            type(self.layer).__name__ if self.layer_name is None else self.layer_name
        )
        params = self.param(
            layer_name, lambda *args: self.layer.init(*args)["params"], inputs
        )
        w = params[self.kernel_name]

        if self.kernel_apply_kwargs is None:
            # By default, we use the implementation in SN-GAN.
            kernel_apply = lambda x: x @ w.reshape(-1, w.shape[-1])
            in_shape = (np.prod(w.shape[:-1]),)
        else:
            # Otherwise, we extract the actual kernel transformation in the input
            # layer. This is useful for Conv2D spectral normalization in
            # [Farzan F. et al., 2019](https://arxiv.org/abs/1811.07457).
            kernel_apply = self.layer.clone(
                **self.kernel_apply_kwargs
            ).bind(  # pylint: disable=not-a-mapping
                {"params": {self.kernel_name: w}}
            )
            # Compute input shape of the kernel operator. This is correct for all
            # linear layers on Flax: Dense, Conv, Embed.
            in_shape = inputs.shape[-w.ndim + 1 : -1] + w.shape[-2:-1]

        initializing = self.is_mutable_collection("params")
        u, v = self._get_singular_vectors(initializing, kernel_apply, in_shape, w.dtype)
        u_hat, v_hat = u.value, v.value
        u_, kernel_transpose = jax.vjp(kernel_apply, v_hat)
        if update_singular_value_estimate and not initializing:
            # Run power iterations using autodiff approach inspired by
            # (Stephan H., 2020)[https://nbviewer.jupyter.org/gist/shoyer/fa9a29fd0880e2e033d7696585978bfc]).
            def scan_body(carry, _):
                u_hat, v_hat, u_ = carry
                (v_,) = kernel_transpose(u_hat)
                v_hat = _l2_normalize(v_)
                u_ = kernel_apply(v_hat)
                u_hat = _l2_normalize(u_)
                return (u_hat, v_hat, u_), None

            (u_hat, v_hat, u_), _ = jax.lax.scan(
                scan_body, (u_hat, v_hat, u_), None, length=self.iteration
            )
            u.value, v.value = u_hat, v_hat

        sigma = jnp.vdot(u_hat, u_)
        # Bound spectral norm by the `norm_multiplier`.
        sigma = jnp.maximum(sigma / self.norm_multiplier, 1.0)
        w_hat = w / jax.lax.stop_gradient(sigma)
        self.sow("intermediates", "w", w_hat)

        # Update params.
        params = flax.core.unfreeze(params)
        params[self.kernel_name] = w_hat
        layer_params = flax.core.freeze({"params": params})
        return self.layer.apply(layer_params, inputs)


class SpectralNormalizationConv2D(SpectralNormalization):
    __doc__ = (
        "Implements spectral normalization for Convolutional layers."
        "See `Generalizable Adversarial Training via Spectral Normalization <https://arxiv.org/abs/1811.07457>`_.\n"
        + "\n".join(SpectralNormalization.__doc__.split("\n")[1:])
    )

    kernel_apply_kwargs: Mapping[str, Any] = flax.core.FrozenDict(
        feature_group_count=1, padding="SAME", use_bias=False
    )


@dataclasses.dataclass
class WithSpectralConv2DNorm:
    """
    Attributes
    ----------
    spectral_norm_iteration: int
        The number of power iterations to estimate weight matrix's singular value.
    spectral_norm_bound: float
        Multiplicative constant to threshold the normalization.
        Usually under normalization, the singular value will converge to this value.
    """
    spectral_norm_iteration: int = 1.
    spectral_norm_bound: float = 0.95

    def spectral_norm(self, layer: nn.Module,  train: bool = False) -> Callable:
        return lambda *args, **kwargs: SpectralNormalizationConv2D(
            layer=layer(*args, **kwargs),
            iteration=self.spectral_norm_iteration,
            norm_multiplier=self.spectral_norm_bound,
            update_singular_value_estimate=train
        )


@dataclasses.dataclass
class WithSpectralNorm:
    """
    Attributes
    ----------
    spectral_norm_iteration: int
        The number of power iterations to estimate weight matrix's singular value.
    spectral_norm_bound: float
        Multiplicative constant to threshold the normalization.
        Usually under normalization, the singular value will converge to this value.
    """
    spectral_norm_iteration: int = 1.
    spectral_norm_bound: float = 0.95

    def spectral_norm(self, layer: nn.Module, train: bool = False) -> Callable:
        return lambda *args, **kwargs: SpectralNormalization(
            layer=layer(*args, **kwargs),
            iteration=self.spectral_norm_iteration,
            norm_multiplier=self.spectral_norm_bound,
            update_singular_value_estimate=train
        )



from fortuna.output_calib_model.output_calib_mixin import WithOutputCalibCheckpointingMixin
from fortuna.training.train_state_repository import TrainStateRepository


class OutputCalibStateRepository(WithOutputCalibCheckpointingMixin, TrainStateRepository):
    pass



from __future__ import annotations

from copy import deepcopy
from itertools import zip_longest
from typing import Iterable, Union, Tuple, Optional, Any, Callable

import numpy as np

from fortuna.typing import Batch, InputData, Targets, Array


class IterableData:
    def __init__(
            self,
            generator,
    ):
        self._g = generator

    def __iter__(self) -> Iterable[Union[Batch, InputData, Targets]]:
        yield from iter(self._g())

    @classmethod
    def from_iterable(cls, iterable) -> IterableData:
        def _inner():
            yield from iterable
        return cls(_inner)

    @classmethod
    def from_callable(cls, fn) -> IterableData:
        def _inner():
            yield from fn()
        return cls(_inner)

    @classmethod
    def from_tf_dataloader(cls, tf_dataloader) -> IterableData:
        def _inner():
            for batch_inputs, batch_targets in tf_dataloader:
                if not isinstance(batch_inputs, dict):
                    batch_inputs = batch_inputs.numpy()
                else:
                    batch_inputs = {k: v.numpy() for k, v in batch_inputs.items()}
                batch_targets = batch_targets.numpy()
                yield batch_inputs, batch_targets
        return cls(_inner)

    @classmethod
    def from_torch_dataloader(cls, torch_dataloader) -> IterableData:
        def _inner():
            for batch_inputs, batch_targets in torch_dataloader:
                if not isinstance(batch_inputs, dict):
                    batch_inputs = batch_inputs.numpy()
                else:
                    batch_inputs = {k: v.numpy() for k, v in batch_inputs.items()}
                batch_targets = batch_targets.numpy()
                yield batch_inputs, batch_targets
        return cls(_inner)

    @classmethod
    def from_array_data(cls, data: Array, batch_size: Optional[int] = None, shuffle: bool = False, prefetch: bool = False, seed: Optional[int] = 0):
            def _inner():
                if shuffle:
                    rng = np.random.default_rng(seed)
                    perm = rng.choice(
                        data.shape[0], data.shape[0], replace=False
                    )
                if batch_size is None:
                    yield data
                else:
                    batches = np.split(
                        data[perm] if shuffle else data,
                        np.arange(batch_size, data.shape[0], batch_size),
                        axis=0,
                    )

                    def make_gen():
                        for batch in batches:
                            yield batch

                    yield from PrefetchedGenerator(make_gen()) if prefetch else make_gen()
            return cls(_inner)

    @classmethod
    def from_batch_array_data(cls, data: Tuple[Array, Array], batch_size: Optional[int] = None, shuffle: bool = False, prefetch: bool = False, seed: Optional[int] = 0):
            def _inner():
                if shuffle:
                    rng = np.random.default_rng(seed)
                    perm = rng.choice(
                        data[0].shape[0], data[0].shape[0], replace=False
                    )
                if batch_size is None:
                    yield data
                else:
                    x_batches = np.split(
                        data[0][perm] if shuffle else data[0],
                        np.arange(batch_size, data[0].shape[0], batch_size),
                        axis=0,
                    )
                    y_batches = np.split(
                        data[1][perm] if shuffle else data[1],
                        np.arange(batch_size, data[1].shape[0], batch_size),
                        axis=0,
                    )

                    def make_gen():
                        for x_batch, y_batch in zip(x_batches, y_batches):
                            yield x_batch, y_batch

                    yield from PrefetchedGenerator(make_gen()) if prefetch else make_gen()
            return cls(_inner)

    @classmethod
    def data_loader_to_inputs_iterable(cls, data_loader) -> IterableData:
        def _inner():
            for inputs, _ in data_loader:
                yield inputs
        return cls(_inner)

    @classmethod
    def data_loader_to_targets_iterable(cls, data_loader) -> IterableData:
        def _inner():
            for _, targets in data_loader:
                yield targets
        return cls(_inner)

    @classmethod
    def inputs_loaders_to_batch_iterable(cls, inputs_loaders, targets) -> IterableData:
        def _inner():
            for all_inputs in zip_longest(*inputs_loaders):
                _targets = [target * np.ones(inputs.shape[0], dtype="int32") for inputs, target in zip(all_inputs, targets) if inputs is not None]
                all_inputs = [inputs for inputs in all_inputs if inputs is not None]
                yield np.concatenate(all_inputs), np.concatenate(_targets)
        return cls(_inner)

    @classmethod
    def transform_data_loader(cls, loader, transform_fn, status) -> IterableData:
        def _inner():
            _status = deepcopy(status)
            for inputs, targets in loader:
                inputs, targets, _status = transform_fn(inputs, targets, _status)
                if inputs is not None and targets is not None and len(inputs) > 0 and len(targets) > 0:
                    if inputs.shape[0] != targets.shape[0]:
                        raise ValueError("The first dimension of transformed inputs and targets must be the same, but "
                                         f"{inputs.shape[0]} and {targets.shape[0]} were found.")
                    yield inputs, targets
        return cls(_inner)

    @classmethod
    def transform_inputs_or_targets_loader(cls, loader, transform_fn, status) -> IterableData:
        def _inner():
            _status = deepcopy(status)
            for data in loader:
                data, _status = transform_fn(data, _status)
                if data is not None and len(data) > 0:
                    yield data
        return cls(_inner)


class PrefetchedGenerator:
    def __init__(self, generator):
        self._batch = generator.__next__()
        self._generator = generator
        self._ready = True

    def __iter__(self):
        return self

    def __next__(self):
        if not self._ready:
            self._prefetch()
        self._ready = False
        return self._batch

    def _prefetch(self):
        if not self._ready:
            self._batch = self._generator.__next__()
            self._ready = True



from fortuna.conformal.batch_mvp import BatchMVPConformalMethod
from fortuna.conformal.classification.base import ConformalClassifier

from typing import Callable, Union, List, Tuple, Optional
from fortuna.typing import Array
from fortuna.data.loader import DataLoader, InputsLoader
import jax.numpy as jnp
from jax import vmap
import numpy as np


class BatchMVPConformalClassifier(BatchMVPConformalMethod, ConformalClassifier):
    def __init__(
            self,
            score_fn: Callable[[Array, Array], Array],
            group_fns: List[Callable[[Array], Array]],
            n_classes: int,
            n_buckets: int = 100
    ):
        """
        This class implements a classification version of BatchMVP
        `[Jung et al., 2022] <https://arxiv.org/abs/2209.15145>`_,
        a multivalid conformal prediction method that satisfies coverage guarantees conditioned on group membership
        and non-conformity threshold.

        Parameters
        ----------
        score_fn: Callable[[Array, Array], Array]
            A score function mapping a batch of inputs and targets to scalar scores, one for each data point. The
            score function represents the degree of non-conformity between inputs and targets. In regression, an
            example of score function is :math:`s(x,y)=|y - h(x)|`, where `h` is an arbitrary regression model.
        group_fns: List[Callable[[Array], Array]]
            A list of group functions, each mapping input data points into boolean arrays which determine whether
            an input belongs to a certain group or not. As an example, suppose that we are interested in obtaining
            marginal coverages guarantee on both negative and positive scalar inputs.
            Then we could define groups functions
            :math:`g_1(x) = x < 0` and :math:`g_1(x) = x > 0`.
            Note that groups can be overlapping, and do not need to cover the full space of inputs.
        n_classes: int
            The number of distinct classes to classify among. The underlying assumption is that the classes are
            identified with an integer from 0 to :code:`n_classes-1`.
        n_buckets: int
            The number of buckets that defines the search space between 0 and 1 that determines the updates of the
            thresholds for the score function.
        """
        super().__init__(score_fn=score_fn, group_fns=group_fns, n_buckets=n_buckets)
        self.n_classes = n_classes

    def conformal_set(
            self,
            val_data_loader: DataLoader,
            test_inputs_loader: InputsLoader,
            error: float = 0.05,
            tol: float = 1e-4,
            n_rounds: int = 1000,
            return_max_calib_error: bool = False,
            test_thresholds: Optional[Array] = None
    ) -> Union[List[List[int]], Tuple[List[List[int]], List[Array]]]:
        """
        Compute a conformal set for each test input.

        Parameters
        ----------
        val_data_loader: DataLoader
            A data loader of validation data.
        test_inputs_loader: InputsLoader
            A loader of test input data points.
        error: float
            A desired coverage error.
        tol: float
            A tolerance for the maximum calibration error.
        n_rounds: int
            The maximum number of updates the algorithm will run for.
        return_max_calib_error: bool
            Whether to return a list of computed maximum calibration error, that is the larger calibration error
            over the different groups.
        test_thresholds: Optional[Array]
            The score thresholds computed over the test data set. These should be the output of
            `BatchMVP.threshold_score`. If provided, they will not be recomputed internally.

        Returns
        -------
        Union[List[List[int]], Tuple[List[List[int]], List[Array]]]
            The computed conformal sets for each test input. Optionally, it returns the maximum calibration errors
            computed during the algorithm.
        """
        if test_thresholds is not None and return_max_calib_error:
            raise ValueError("If `test_thresholds` is given, `return_max_calib_error` cannot be returned.")
        if test_thresholds is None:
            outs = self.threshold_score(
                val_data_loader=val_data_loader,
                test_inputs_loader=test_inputs_loader,
                error=error,
                tol=tol,
                n_rounds=n_rounds,
                return_max_calib_error=return_max_calib_error
            )
            if return_max_calib_error:
                test_thresholds, max_calib_errors = outs
            else:
                test_thresholds = outs

        c = 0
        all_ys = jnp.arange(self.n_classes)
        all_bools = []
        for inputs in test_inputs_loader:
            batch_thresholds = test_thresholds[c:c+inputs.shape[0]]
            all_bools.append(vmap(lambda y: self.score_fn(inputs, y) <= batch_thresholds, out_axes=1)(all_ys))
            c += inputs.shape[0]
        all_bools = jnp.concatenate(all_bools, axis=0)

        sizes = np.sum(all_bools, 1)
        sets = np.zeros(c, dtype=object)
        for s in np.unique(sizes):
            idx = jnp.where(sizes == s)[0]
            sets[idx] = np.nonzero(all_bools[idx])[1].reshape(len(idx), s).tolist()
        sets = sets.tolist()

        if return_max_calib_error:
            return sets, max_calib_errors
        return sets



from fortuna.typing import Batch, Params, Mutable, CalibMutable, CalibParams, Outputs, Targets
from typing import List, Optional, Tuple, Union, Any, Callable
from jax._src.prng import PRNGKeyArray
import jax.numpy as jnp
from fortuna.utils.random import WithRNG
from fortuna.likelihood.base import Likelihood


class Loss(WithRNG):
    def __init__(self, likelihood: Likelihood, loss_fn: Callable[[Outputs, Targets], jnp.ndarray]):
        self.likelihood = likelihood
        self.loss_fn = loss_fn

    def __call__(
        self,
        params: Params,
        batch: Batch,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        return_aux: Optional[List[str]] = None,
        train: bool = False,
        outputs: Optional[jnp.ndarray] = None,
        rng: Optional[PRNGKeyArray] = None,
        **kwargs
    ) -> Union[jnp.ndarray, Tuple[jnp.ndarray, Any]]:
        if return_aux is None:
            return_aux = []
        supported_aux = ["outputs", "mutable", "calib_mutable"]
        unsupported_aux = [s for s in return_aux if s not in supported_aux]
        if sum(unsupported_aux) > 0:
            raise AttributeError(
                """The auxiliary objects {} is unknown. Please make sure that all elements of `return_aux` 
                            belong to the following list: {}""".format(
                    unsupported_aux, supported_aux
                )
            )
        if train and outputs is not None:
            raise ValueError(
                """When `outputs` is available, `train` must be set to `False`."""
            )
        if "mutable" in return_aux and outputs is not None:
            raise ValueError(
                """When `outputs` is available, `return_aux` cannot contain 'mutable'`."""
            )
        if not train and "mutable" in return_aux:
            raise ValueError(
                "Returning an auxiliary mutable is supported only during training. Please either set `train` to "
                "`True`, or remove 'mutable' from `return_aux`."
            )
        if "mutable" in return_aux and mutable is None:
            raise ValueError(
                "In order to be able to return an auxiliary mutable, an initial mutable must be passed as `mutable`. "
                "Please either remove 'mutable' from `return_aux`, or pass an initial mutable as `mutable`."
            )
        if "mutable" not in return_aux and mutable is not None and train is True:
            raise ValueError(
                """You need to add `mutable` to `return_aux`. When you provide a (not null) `mutable` variable during 
                training, that variable will be updated during the forward pass."""
            )

        inputs, targets = batch
        if outputs is None:
            outs = self.likelihood.model_manager.apply(
                params,
                inputs,
                train=train,
                mutable=mutable,
                rng=rng,
            )
            if "mutable" in return_aux:
                outputs, aux = outs
                mutable = aux["mutable"]
            else:
                outputs = outs

        aux = dict()
        if self.likelihood.output_calib_manager is not None:
            outs = self.likelihood.output_calib_manager.apply(
                params=calib_params["output_calibrator"]
                if calib_params is not None
                else None,
                mutable=calib_mutable["output_calibrator"]
                if calib_mutable is not None
                else None,
                outputs=outputs,
                calib="calib_mutable" in return_aux,
            )
        if (
            calib_mutable is not None
            and calib_mutable["output_calibrator"] is not None
            and "calib_mutable" in return_aux
        ):
            outputs, aux["calib_mutable"] = outs
            aux["calib_mutable"] = dict(output_calibrator=aux["calib_mutable"])
        else:
            outputs = outs
            if "calib_mutable" in return_aux:
                aux["calib_mutable"] = dict(output_calibrator=None)

        if "outputs" in return_aux:
            aux["outputs"] = outputs
        if "mutable" in return_aux:
            aux["mutable"] = mutable

        loss = self.loss_fn(outputs, targets)
        if len(return_aux) > 0:
            return loss, aux
        return loss



from fortuna.prob_model.posterior.normalizing_flow.advi import ADVI_NAME
from fortuna.prob_model.posterior.normalizing_flow.normalizing_flow_trainer import \
    NormalizingFlowTrainer
from fortuna.utils.nested_dicts import nested_set, nested_pair
from fortuna.prob_model.posterior.normalizing_flow.advi.advi_state import ADVIState
from fortuna.typing import Path, Params
from typing import Optional
from flax.core import FrozenDict
import jax
from fortuna.training.trainer import MultiDeviceMixin, JittedMixin
from jax.tree_util import tree_map


class ADVITrainer(NormalizingFlowTrainer):
    def __str__(self):
        return ADVI_NAME

    def save_checkpoint(
            self,
            state: ADVIState,
            save_checkpoint_dir: Path,
            keep: int = 1,
            force_save: bool = False,
            prefix: str = "checkpoint_",
    ) -> None:
        state = state.replace(params=self._unravel_params(state.params, self._all_params))
        super().save_checkpoint(
            state,
            save_checkpoint_dir,
            keep,
            force_save,
            prefix
        )

    def _unravel_params(
            self,
            rav_params: FrozenDict,
            all_params: Optional[Params] = None,
    ) -> Params:
        if self._which_params is not None:
            means = tuple([_unravel(
                rav_params["mean"][self._indices[i]:self._indices[i + 1]]) for i, _unravel in enumerate(self._unravel)])
            log_stds = tuple([_unravel(rav_params["log_std"][self._indices[i]:self._indices[i + 1]]) for i, _unravel in enumerate(self._unravel)])

            all_params = all_params.unfreeze()
            all_params = nested_set(all_params, self._which_params, means)
            all_params = nested_pair(
                all_params,
                self._which_params,
                log_stds,
                ("mean", "log_std"),
            )
            return FrozenDict(all_params)

        params = self._unravel(rav_params["mean"]).unfreeze()
        log_stds = self._unravel(rav_params["log_std"])
        for k, v in params.items():
            params[k] = {"params": dict(mean=v["params"], log_std=log_stds[k]["params"])}
        return FrozenDict(params)

    def on_train_end(self, state: ADVIState) -> ADVIState:
        self.save_checkpoint(
            state,
            save_checkpoint_dir=self.save_checkpoint_dir,
            keep=self.keep_top_n_checkpoints,
            force_save=True,
        )
        return state.replace(params=self._unravel_params(state.params, self._all_params))


class JittedADVITrainer(JittedMixin, ADVITrainer):
    pass


class MultiDeviceADVITrainer(MultiDeviceMixin, ADVITrainer):
    def on_train_end(self, state: ADVIState) -> ADVIState:
        self.save_checkpoint(
            state,
            save_checkpoint_dir=self.save_checkpoint_dir,
            keep=self.keep_top_n_checkpoints,
            force_save=True,
        )
        state = jax.device_get(tree_map(lambda x: x[0], state))
        return state.replace(params=self._unravel_params(state.params, self._all_params))



from fortuna.typing import Array
import jax.numpy as jnp
import jax.scipy as jsp
import jax


def cross_entropy_loss_fn(
        outputs: Array,
        targets: Array,
) -> jnp.ndarray:
    """
    A cross-entropy loss function. Check `here <https://en.wikipedia.org/wiki/Cross_entropy>`_ for reference.

    Parameters
    ----------
    outputs: Array
        Model outputs to be passed to the loss.
    targets: Array
        Target data points.

    Returns
    -------
    jnp.ndarray
        The focal loss evaluation.
    """
    targets = jax.nn.one_hot(targets, outputs.shape[-1])
    return jnp.mean(jsp.special.logsumexp(outputs, -1) - jnp.sum(targets * outputs, -1))
